using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Text;
using USB = PICkit2V2.USB;
using KONST = PICkit2V2.Constants;
using UTIL = PICkit2V2.Utilities;
using System.IO;
using File = System.IO.File;

namespace PICkit2V2
{
    public class PICkitFunctions
    {
        public static String FirmwareVersion = "NA";
        public static String DeviceFileVersion = "NA";
        public static DeviceFile DevFile = new DeviceFile();
        public static DeviceData DeviceBuffers;
        public static byte[] Usb_write_array = new byte[KONST.PACKET_SIZE];
        public static byte[] Usb_read_array = new byte[KONST.PACKET_SIZE]; 
        public static int ActivePart = 0;
        public static int ActiveRev = 0;
        public static uint LastDeviceID = 0;
    
        private static IntPtr usbReadHandle = IntPtr.Zero;
        private static IntPtr usbWriteHandle = IntPtr.Zero;
        private static int[] familySearchTable; // index is search priority, value is family array index.
        private static bool vddOn = false;
        private static float vddLastSet = 3.3F;  // needed when a family VPP=VDD (PIC18J, PIC24, etc.)
        private static bool targetSelfPowered = false;
        private static uint scriptBufferChecksum = 0;
        private static int lastFoundPart = 0;
        
        public static void ForcePICkitPowered()
        {
            targetSelfPowered = false;
        }

        public static void ForceTargetPowered()
        {
            targetSelfPowered = true;
        }        

        public static void ReadConfigOutsideProgMem()
        {
            RunScript(KONST.PROG_ENTRY, 1);
            RunScript(KONST.CONFIG_RD, 1);
            UploadData();
            RunScript(KONST.PROG_EXIT, 1);
            int configWords = DevFile.PartsList[ActivePart].ConfigWords;
            int bufferIndex = 2;                    // report starts on index 1, which is #bytes uploaded.
            for (int word = 0; word < configWords; word++)
            {
                uint config = (uint)Usb_read_array[bufferIndex++];
                config |= (uint)Usb_read_array[bufferIndex++] << 8;
                if (DevFile.Families[GetActiveFamily()].ProgMemShift > 0)
                {
                    config = (config >> 1) & DevFile.Families[GetActiveFamily()].BlankValue;
                }
                DeviceBuffers.ConfigWords[word] = config;
            }        
        }
        
        public static void ReadBandGap()
        {
            RunScript(KONST.PROG_ENTRY, 1);
            RunScript(KONST.CONFIG_RD, 1);
            UploadData();
            RunScript(KONST.PROG_EXIT, 1);
            int configWords = DevFile.PartsList[ActivePart].ConfigWords;
            uint config = (uint)Usb_read_array[2];
            config |= (uint)Usb_read_array[3] << 8;
            if (DevFile.Families[GetActiveFamily()].ProgMemShift > 0)
            {
                config = (config >> 1) & DevFile.Families[GetActiveFamily()].BlankValue;
            }
            DeviceBuffers.BandGap = config & DevFile.PartsList[ActivePart].BandGapMask;
        }
        
        public static void WriteConfigOutsideProgMem(bool codeProtect, bool dataProtect)
        {
            int configWords = DevFile.PartsList[ActivePart].ConfigWords;
            byte[] configBuffer = new byte[configWords * 2];
            
            if (DevFile.PartsList[ActivePart].BandGapMask > 0)
            {
                DeviceBuffers.ConfigWords[0] &= ~DevFile.PartsList[ActivePart].BandGapMask;
                DeviceBuffers.ConfigWords[0] |= DeviceBuffers.BandGap;
            }

            RunScript(KONST.PROG_ENTRY, 1);

            if (DevFile.PartsList[ActivePart].ConfigWrPrepScript > 0)
            {
                DownloadAddress3(0);
                RunScript(KONST.CONFIG_WR_PREP, 1);
            }    

            for (int i = 0, j = 0; i < configWords; i++)
            {
                uint configWord = DeviceBuffers.ConfigWords[i];
                if (i == DevFile.PartsList[ActivePart].CPConfig - 1)
                {
                    if (codeProtect)
                    {
                        configWord &= (uint)~DevFile.PartsList[ActivePart].CPMask;
                    }
                    if (dataProtect)
                    {
                        configWord &= (uint)~DevFile.PartsList[ActivePart].DPMask;
                    }
                }
                if (DevFile.Families[GetActiveFamily()].ProgMemShift > 0)
                {
                    configWord = configWord << 1;
                }
                configBuffer[j++] = (byte)(configWord & 0xFF);
                configBuffer[j++] = (byte)((configWord >> 8) & 0xFF);
            }
            DataClrAndDownload(configBuffer, 0);
            RunScript(KONST.CONFIG_WR, 1);
            RunScript(KONST.PROG_EXIT, 1);        
        
        }

        public static bool ReadOSSCAL()
        {
            if (RunScript(KONST.PROG_ENTRY, 1))
            {
                if (DownloadAddress3((int)(DevFile.PartsList[ActivePart].ProgramMem - 1)))
                {
                    if (RunScript(KONST.OSSCAL_RD, 1))
                    {
                        if (UploadData())
                        {
                            if (RunScript(KONST.PROG_EXIT, 1))
                            {
                                DeviceBuffers.OSCCAL = (uint)(Usb_read_array[2] + (Usb_read_array[3] * 256));
                                if (DevFile.Families[GetActiveFamily()].ProgMemShift > 0)
                                {
                                    DeviceBuffers.OSCCAL >>= 1;
                                }
                                DeviceBuffers.OSCCAL &= DevFile.Families[GetActiveFamily()].BlankValue;
                                //DeviceBuffers.OSCCAL = 0xc00;
                                return true;
                            }
                        }
                    }
                }
            }
            return false;
        }

        public static bool WriteOSSCAL()
        {
            if (RunScript(KONST.PROG_ENTRY, 1))
            {
                uint calWord = DeviceBuffers.OSCCAL;
                uint calAddress = DevFile.PartsList[ActivePart].ProgramMem - 1;
                if (DevFile.Families[GetActiveFamily()].ProgMemShift > 0)
                {
                    calWord <<= 1;
                }
                byte[] addressData = new byte[5];
                addressData[0] = (byte)(calAddress & 0xFF);
                addressData[1] = (byte)((calAddress >> 8) & 0xFF);
                addressData[2] = (byte)((calAddress >> 16) & 0xFF);
                addressData[3] = (byte)(calWord & 0xFF);
                addressData[4] = (byte)((calWord >> 8) & 0xFF);
                DataClrAndDownload(addressData, 0);
                if (RunScript(KONST.OSSCAL_WR, 1))
                {
                    if (RunScript(KONST.PROG_EXIT, 1))
                    {
                        return true;
                    }
                }
            }
            return false;
        }        

        public static KONST.PICkit2PWR CheckTargetPower(ref float vdd, ref float vpp)
        {
            if (vddOn)  // if VDD on, can't check for self-powered target
            {
                return KONST.PICkit2PWR.vdd_on;
            }
            
            if (readPICkitVoltages(ref vdd, ref vpp))
            {
                if (vdd > KONST.VddThresholdForSelfPoweredTarget)
                {
                    targetSelfPowered = true;
                    SetVDDVoltage(vdd);                     // set VDD to target level
                    return KONST.PICkit2PWR.selfpowered;
                }
                targetSelfPowered = false;
                return KONST.PICkit2PWR.unpowered;
            }
            targetSelfPowered = false;
            return KONST.PICkit2PWR.no_response;
        }
               
        public static int GetActiveFamily()
        {
            return DevFile.PartsList[ActivePart].Family;
        }

        public static void SetActiveFamily(int family)
        {
            ActivePart = 0;
            lastFoundPart = 0;
            DevFile.PartsList[ActivePart].Family = (ushort)family;
            // Set up the buffers for unsupported part (else they remain last selected part)
            ResetBuffers();
        }
        
        public static bool SetVDDVoltage(float voltage)
        {
            vddLastSet = voltage;

            ushort ccpValue = (ushort)(voltage * 32F + 10.5F);
            ccpValue <<= 6;
            byte vFault = (byte) (((0.85F * voltage) / 5F) * 255F);
            
            byte[] commandArray = new byte[4];
            commandArray[0] = KONST.SETVDD;
            commandArray[1] = (byte) (ccpValue & 0xFF);
            commandArray[2] = (byte) (ccpValue / 256);
            commandArray[3] = vFault;
            return writeUSB(commandArray);
        }
        
        public static bool VddOn()
        {
                byte [] commandArray = new byte[4];
                commandArray[0] = KONST.EXECUTE_SCRIPT;
                commandArray[1] = 2;
                commandArray[2] = KONST._VDD_GND_OFF;
                if (targetSelfPowered)
                {   // don't turn on VDD if self-powered!
                    commandArray[3] = KONST._VDD_OFF;
                }
                else
                {
                    commandArray[3] = KONST._VDD_ON;
                }
                bool result = writeUSB(commandArray);
                if (result)
                {
                    vddOn = true;
                    return true;
                }
                return result;    
        }

        public static bool VddOff()
        {
            byte[] commandArray = new byte[4];
            commandArray[0] = KONST.EXECUTE_SCRIPT;
            commandArray[1] = 2;
            commandArray[2] = KONST._VDD_OFF;
            if (targetSelfPowered)
            { // don't ground VDD if self-powered target
                commandArray[3] = KONST._VDD_GND_OFF;
            }
            else
            {
                commandArray[3] = KONST._VDD_GND_ON;
            }
            bool result = writeUSB(commandArray);
            if (result)
            {
                vddOn = false;
                return true;
            }
            return result;
        }
        
        public static bool SetProgrammingSpeed(byte speed)
        {
            byte[] commandArray = new byte[4];
            commandArray[0] = KONST.EXECUTE_SCRIPT;
            commandArray[1] = 2;
            commandArray[2] = KONST._SET_ICSP_SPEED; 
            commandArray[3] = speed;
            return writeUSB(commandArray);
        }

        public static bool ResetPICkit2()
        {
            byte[] commandArray = new byte[1];
            commandArray[0] = KONST.RESET;
            return writeUSB(commandArray);
        }

        public static bool EnterBootloader()
        {
            byte[] commandArray = new byte[1];
            commandArray[0] = KONST.ENTER_BOOTLOADER;
            return writeUSB(commandArray);
        } 

        public static bool VerifyBootloaderMode()
        {
            byte[] commandArray = new byte[1];
            commandArray[0] = KONST.FIRMWARE_VERSION;
            if (writeUSB(commandArray))
            {
                if (readUSB())
                {
                    if (Usb_read_array[1] == 118) // ASCII 'B'
                    {
                        return true;
                    }
                    return false;
                }
                return false;
            }
            return false;
        }

        public static bool BL_EraseFlash()
        {
            byte[] commandArray = new byte[5];
            commandArray[0] = KONST.ERASEFWFLASH;
            commandArray[1] = 0xC0;
            commandArray[2] = 0x00;
            commandArray[3] = 0x20;
            commandArray[4] = 0x00;
            if (writeUSB(commandArray))
            {
                commandArray[3] = 0x50;
                return writeUSB(commandArray);
            }
            return false;
        }

        public static bool BL_WriteFlash(byte[] payload)
        {
            byte[] commandArray = new byte[2+35];
            commandArray[0] = KONST.WRITEFWFLASH;
            commandArray[1] = 32;
            for (int i=0; i< 35; i++)
            {
                commandArray[2+i] = payload[i];
            }
            return writeUSB(commandArray);

        }
        
        public static bool BL_WriteFWLoadedKey()
        {
            byte[] flashWriteData = new byte[3 + 32];  // 3 address bytes plus 32 data bytes.
            
            flashWriteData[0] = 0xE0;
            flashWriteData[1] = 0x7F;
            flashWriteData[2] = 0x00;   // Address = 0x007FE0
            for (int i = 3; i < flashWriteData.Length; i++)
            {
                flashWriteData[i] = 0xFF;
            }
            flashWriteData[flashWriteData.Length - 2] = 0x55;
            flashWriteData[flashWriteData.Length - 1] = 0x55;
            return BL_WriteFlash(flashWriteData);
        }

        public static bool BL_ReadFlash16(int address)
        {
            byte[] commandArray = new byte[5];
            commandArray[0] = KONST.READFWFLASH;
            commandArray[1] = 16;
            commandArray[2] = (byte)(address & 0xFF);
            commandArray[3] = (byte)((address >> 8) & 0xFF);
            commandArray[4] = 0x00;
            if (writeUSB(commandArray))
            {
                return readUSB();
            }
            return false;

        }

        public static bool BL_Reset()
        {
            byte[] commandArray = new byte[1];
            commandArray[0] = KONST.RESETFWDEVICE;

            return writeUSB(commandArray);

        }          
        
        public static bool ButtonPressed()
        {
            ushort status = readPkStatus();
            if ((status & 0x0040) == 0x0040)
            {
                return true;
            }
            return false;
        }                 
        
        public static KONST.PICkit2PWR PowerStatus()
        {
            ushort status = readPkStatus();
            if (status == 0xFFFF)
            {
                return KONST.PICkit2PWR.no_response;
            }
            if ((status & 0x0030) == 0x0030)
            {
                vddOn = false;
                return KONST.PICkit2PWR.vddvpperrors;
            }
            if ((status & 0x0020) == 0x0020)
            {
                vddOn = false;
                return KONST.PICkit2PWR.vpperror;
            }
            if ((status & 0x0010) == 0x0010)
            {
                vddOn = false;
                return KONST.PICkit2PWR.vdderror;
            }
            if ((status & 0x0002) == 0x0002)
            {
                vddOn = true;
                return KONST.PICkit2PWR.vdd_on;
            }
            vddOn = false;
            return KONST.PICkit2PWR.vdd_off;
            
        }
       
     
        public static KONST.PICkit2USB DetectPICkit2Device()
        {
            bool result = USB.Find_This_Device(KONST.MChipVendorID, KONST.Pk2DeviceID,
                                 0, ref usbReadHandle, ref usbWriteHandle);
            if (result)
            {
                //Read firmware version
                byte [] commandArray = new byte[1];
                commandArray[0] = KONST.FIRMWARE_VERSION;
                result = writeUSB(commandArray);

                if (result)
                {
                    // read response
                    if (readUSB())
                    {
                        // create a version string
                        FirmwareVersion = string.Format("{0:D1}.{1:D2}.{2:D2}", Usb_read_array[1],
                                    Usb_read_array[2], Usb_read_array[3]);
                        // check for minimum supported version
                        if (Usb_read_array[1] == KONST.FWVerMajorReq)
                        {
                            if (((Usb_read_array[2] == KONST.FWVerMinorReq) 
                                && (Usb_read_array[3] >= KONST.FWVerDotReq))
                                || (Usb_read_array[2] > KONST.FWVerMinorReq))
                            {
                                return KONST.PICkit2USB.found;
                            }
                        }
                        if (Usb_read_array[1] == 118)
                        {
                            return KONST.PICkit2USB.bootloader;
                        
                        }
                        return KONST.PICkit2USB.firmwareInvalid;
                    }
                    return KONST.PICkit2USB.readError;
                }
                return KONST.PICkit2USB.writeError;
            }
            return KONST.PICkit2USB.notFound;

        }

        public static bool ReadDeviceFile()
        {
            string DeviceFileName = "PK2DeviceFile.dat";

            bool fileExists = File.Exists(DeviceFileName);
            if (fileExists)
            {
                try
                {
                    FileStream fsDevFile = File.Open(DeviceFileName, FileMode.Open);
                    using (BinaryReader binRead = new BinaryReader(fsDevFile))
                    {
                        //
                        DevFile.Info.VersionMajor = binRead.ReadInt32();
                        DevFile.Info.VersionMinor = binRead.ReadInt32();
                        DevFile.Info.VersionDot = binRead.ReadInt32();
                        DevFile.Info.VersionNotes = binRead.ReadString();
                        DevFile.Info.NumberFamilies = binRead.ReadInt32();
                        DevFile.Info.NumberParts = binRead.ReadInt32();
                        DevFile.Info.NumberScripts = binRead.ReadInt32();
                        DevFile.Info.Compatibility  = binRead.ReadByte();
                        DevFile.Info.UNUSED1A = binRead.ReadByte();
                        DevFile.Info.UNUSED1B = binRead.ReadUInt16();
                        DevFile.Info.UNUSED2 = binRead.ReadUInt32();
                        // create a version string
                        DeviceFileVersion = string.Format("{0:D1}.{1:D2}.{2:D2}", DevFile.Info.VersionMajor,
                                         DevFile.Info.VersionMinor, DevFile.Info.VersionDot);
                        //
                        // Declare arrays
                        //
                        DevFile.Families = new DeviceFile.DeviceFamilyParams[DevFile.Info.NumberFamilies];
                        DevFile.PartsList = new DeviceFile.DevicePartParams[DevFile.Info.NumberParts];
                        DevFile.Scripts = new DeviceFile.DeviceScripts[DevFile.Info.NumberScripts];
                        //
                        // now read all families if they are there
                        //
                        for (int l_x = 0; l_x < DevFile.Info.NumberFamilies; l_x++)
                        {
                            DevFile.Families[l_x].FamilyID = binRead.ReadUInt16();
                            DevFile.Families[l_x].FamilyType = binRead.ReadUInt16();
                            DevFile.Families[l_x].SearchPriority = binRead.ReadUInt16();
                            DevFile.Families[l_x].FamilyName = binRead.ReadString();
                            DevFile.Families[l_x].ProgEntryScript = binRead.ReadUInt16();
                            DevFile.Families[l_x].ProgExitScript = binRead.ReadUInt16();
                            DevFile.Families[l_x].ReadDevIDScript = binRead.ReadUInt16();
                            DevFile.Families[l_x].DeviceIDMask = binRead.ReadUInt32();
                            DevFile.Families[l_x].BlankValue = binRead.ReadUInt32();
                            DevFile.Families[l_x].BytesPerLocation = binRead.ReadByte();
                            DevFile.Families[l_x].AddressIncrement = binRead.ReadByte();
                            DevFile.Families[l_x].PartDetect = binRead.ReadBoolean();
                            DevFile.Families[l_x].UNUSED1 = binRead.ReadUInt32();
                            DevFile.Families[l_x].EEMemBytesPerWord = binRead.ReadByte();
                            DevFile.Families[l_x].EEMemAddressIncrement = binRead.ReadByte();
                            DevFile.Families[l_x].UserIDHexBytes = binRead.ReadByte();
                            DevFile.Families[l_x].UserIDBytes = binRead.ReadByte();
                            DevFile.Families[l_x].ProgMemHexBytes = binRead.ReadByte();
                            DevFile.Families[l_x].EEMemHexBytes = binRead.ReadByte();
                            DevFile.Families[l_x].ProgMemShift = binRead.ReadByte();
                            DevFile.Families[l_x].TestMemoryStart = binRead.ReadUInt32();
                            DevFile.Families[l_x].TestMemoryLength = binRead.ReadUInt16();
                            DevFile.Families[l_x].Vpp = binRead.ReadSingle();
                        }
                        // Create family search table based on priority
                        familySearchTable = new int[DevFile.Info.NumberFamilies];
                        for (int familyIdx = 0; familyIdx < DevFile.Info.NumberFamilies; familyIdx++)
                        {
                            familySearchTable[DevFile.Families[familyIdx].SearchPriority] = familyIdx;
                        }
                        //
                        // now read all parts if they are there
                        //
                        for (int l_x = 0; l_x < DevFile.Info.NumberParts; l_x++)
                        {
                            DevFile.PartsList[l_x].PartName = binRead.ReadString();
                            DevFile.PartsList[l_x].Family = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].DeviceID = binRead.ReadUInt32();
                            DevFile.PartsList[l_x].ProgramMem = binRead.ReadUInt32();
                            DevFile.PartsList[l_x].EEMem = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].EEAddr = binRead.ReadUInt32();
                            DevFile.PartsList[l_x].ConfigWords = binRead.ReadByte();
                            DevFile.PartsList[l_x].ConfigAddr = binRead.ReadUInt32();
                            DevFile.PartsList[l_x].UserIDWords = binRead.ReadByte();
                            DevFile.PartsList[l_x].UserIDAddr = binRead.ReadUInt32();
                            DevFile.PartsList[l_x].BandGapMask = binRead.ReadUInt32();
                            // Init config arrays
                            DevFile.PartsList[l_x].ConfigMasks = new ushort[KONST.NumConfigMasks];
                            DevFile.PartsList[l_x].ConfigBlank = new ushort[KONST.NumConfigMasks];
                            for (int l_index = 0; l_index < KONST.NumConfigMasks; l_index++)
                            {
                                DevFile.PartsList[l_x].ConfigMasks[l_index] = binRead.ReadUInt16();
                            }
                            for (int l_index = 0; l_index < KONST.NumConfigMasks; l_index++)
                            {
                                DevFile.PartsList[l_x].ConfigBlank[l_index] = binRead.ReadUInt16();
                            }
                            DevFile.PartsList[l_x].CPMask = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].CPConfig = binRead.ReadByte();
                            DevFile.PartsList[l_x].OSSCALSave = binRead.ReadBoolean();
                            DevFile.PartsList[l_x].VerifyStop = binRead.ReadUInt32();
                            DevFile.PartsList[l_x].VddMin = binRead.ReadSingle();
                            DevFile.PartsList[l_x].VddMax = binRead.ReadSingle();
                            DevFile.PartsList[l_x].VddErase = binRead.ReadSingle();
                            DevFile.PartsList[l_x].DefaultProgSpeed = binRead.ReadByte();
                            DevFile.PartsList[l_x].ChipEraseScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].ProgMemAddrSetScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].ProgMemAddrBytes = binRead.ReadByte();
                            DevFile.PartsList[l_x].ProgMemRdScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].ProgMemRdWords = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].EERdPrepScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].EERdScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].EERdLocations = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].UserIDRdPrepScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].UserIDRdScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].ConfigRdPrepScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].ConfigRdScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].ProgMemWrPrepScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].ProgMemWrScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].ProgMemWrWords = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].ProgMemPanelBufs = binRead.ReadByte();
                            DevFile.PartsList[l_x].ProgMemPanelOffset = binRead.ReadUInt32();
                            DevFile.PartsList[l_x].EEWrPrepScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].EEWrScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].EEWrLocations = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].UserIDWrPrepScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].UserIDWrScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].ConfigWrPrepScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].ConfigWrScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].OSCCALRdScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].OSCCALWrScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].DPMask = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].WriteCfgOnErase = binRead.ReadBoolean();
                            DevFile.PartsList[l_x].UNUSED1 = binRead.ReadByte();
                            DevFile.PartsList[l_x].UNUSED2 = binRead.ReadUInt32();
                            DevFile.PartsList[l_x].UNUSED3 = binRead.ReadUInt32();
                            DevFile.PartsList[l_x].UNUSED4 = binRead.ReadUInt32();
                            DevFile.PartsList[l_x].ProgMemEraseScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].EEMemEraseScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].ConfigMemEraseScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].reserved1EraseScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].reserved2EraseScript = binRead.ReadUInt16(); 
                            DevFile.PartsList[l_x].TestMemoryRdScript= binRead.ReadUInt16();
	                        DevFile.PartsList[l_x].TestMemoryRdWords = binRead.ReadUInt16();
	                        DevFile.PartsList[l_x].TestMemoryWrScript = binRead.ReadUInt16();
	                        DevFile.PartsList[l_x].TestMemoryWrWords = binRead.ReadUInt16();
	                        DevFile.PartsList[l_x].ExportToMPLAB = binRead.ReadBoolean();
	                        DevFile.PartsList[l_x].DebugHaltScript = binRead.ReadUInt16();
	                        DevFile.PartsList[l_x].DebugRunScript = binRead.ReadUInt16();
	                        DevFile.PartsList[l_x].DebugStatusScript = binRead.ReadUInt16();
	                        DevFile.PartsList[l_x].DebugReadExecVerScript = binRead.ReadUInt16();
	                        DevFile.PartsList[l_x].DebugSingleStepScript = binRead.ReadUInt16();
	                        DevFile.PartsList[l_x].DebugBulkWrDataScript = binRead.ReadUInt16();
	                        DevFile.PartsList[l_x].DebugBulkRdDataScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].DebugWriteVectorScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].DebugReadVectorScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].DebugRowEraseScript = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].DebugReserved4Script = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].DebugReserved5Script = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].DebugReserved6Script = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].DebugReserved7Script = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].DebugReserved8Script = binRead.ReadUInt16();
                            DevFile.PartsList[l_x].DebugReserved9Script = binRead.ReadUInt16();                                                       
                        }
                        //
                        // now read all scripts if they are there
                        //                    
                        for (int l_x = 0; l_x < DevFile.Info.NumberScripts; l_x++)
                        {
                            DevFile.Scripts[l_x].ScriptNumber = binRead.ReadUInt16();
                            DevFile.Scripts[l_x].ScriptName = binRead.ReadString();
                            DevFile.Scripts[l_x].ScriptVersion = binRead.ReadUInt16();
                            DevFile.Scripts[l_x].UNUSED1 = binRead.ReadUInt32();
                            DevFile.Scripts[l_x].ScriptLength = binRead.ReadUInt16();
                            // init script array
                            DevFile.Scripts[l_x].Script = new ushort[DevFile.Scripts[l_x].ScriptLength];
                            for (int l_index = 0; l_index < DevFile.Scripts[l_x].ScriptLength; l_index++)
                            {
                                DevFile.Scripts[l_x].Script[l_index] = binRead.ReadUInt16();
                            }
                            DevFile.Scripts[l_x].Comment = binRead.ReadString();
                        }

                        binRead.Close();
                    }
                    fsDevFile.Close();
                }
                catch
                {
                    return false;
                } 
                return true;
            }
            else
            {
                return false;
            }

        }

        public static bool DetectDevice(int familyIndex, bool resetOnNotFound, bool keepVddOn)
        {
            // Detect a device in the given Family of familyIndex, or all families
            
            if (familyIndex == KONST.SEARCH_ALL_FAMILIES)
            {
                // when searching all families, set Vdd = 3.3v
                if (!targetSelfPowered)
                { //but not if self-powered target
                    SetVDDVoltage(3.3F);
                }
                
                for (int searchIndex = 0; searchIndex < DevFile.Families.Length; searchIndex++)
                {
                    if (DevFile.Families[familySearchTable[searchIndex]].PartDetect)
                    {
                        if (searchDevice(familySearchTable[searchIndex], true, keepVddOn))   
                        // 0 = no supported part found
                        {
                            return true;
                        }
                    }
                }
                return false; // no supported part found in any family
            }
            else
            {
                if (DevFile.Families[familyIndex].PartDetect)
                {
                    if (searchDevice(familyIndex, resetOnNotFound, keepVddOn))
                    {
                        return true;
                    }
                    return false;
                }
                else
                {
                    return true;    // don't fail unsearchable families like baseline.
                }
            }
            
        
        }
        
        public static int FindLastUsedInBuffer(uint[] bufferToSearch, uint blankValue,
                                                int startIndex)
        {   // go backawards from the start entry to find the last non-blank entry
            for (int index = startIndex; index > 0; index--)
            {
                if (bufferToSearch[index] != blankValue)
                {
                    return index;
                }
            }
            
            return 0;          
        }

        // Louis - execute an arbitrary script (upload)
        // It seems like we have to expose the PICkit's inner behavior...
        public static bool ExecuteScriptUploadNoLen(byte[] script)
        {
            byte[] commandArray = new byte[4 + script.Length];
            commandArray[0] = KONST.CLR_UPLOAD_BUFFER;
            commandArray[1] = KONST.EXECUTE_SCRIPT;
            commandArray[2] = (byte)script.Length;
            for (int n = 0; n < script.Length; n++)
            {
                commandArray[3 + n] = script[n];
            }
            commandArray[3 + script.Length] = KONST.UPLOAD_DATA_NOLEN;
            bool result = writeUSB(commandArray);
            if (result)
                result = readUSB();

            return result;
        }

        public static bool RunScriptUploadNoLen(int script, int repetitions)
        {
            // IMPORTANT NOTE: THIS ALWAYS CLEARS THE UPLOAD BUFFER FIRST!

            byte[] commandArray = new byte[5];
            commandArray[0] = KONST.CLR_UPLOAD_BUFFER;
            commandArray[1] = KONST.RUN_SCRIPT;
            commandArray[2] = (byte)script;
            commandArray[3] = (byte)repetitions;
            commandArray[4] = KONST.UPLOAD_DATA_NOLEN;
            bool result = writeUSB(commandArray);
            if (result)
            {
                result = readUSB();
            }
            return result;
        }

        public static bool RunScriptUploadNoLen2(int script, int repetitions)
        {
            // IMPORTANT NOTE: THIS ALWAYS CLEARS THE UPLOAD BUFFER FIRST!

            byte[] commandArray = new byte[6];
            commandArray[0] = KONST.CLR_UPLOAD_BUFFER;
            commandArray[1] = KONST.RUN_SCRIPT;
            commandArray[2] = (byte)script;
            commandArray[3] = (byte)repetitions;
            commandArray[4] = KONST.UPLOAD_DATA_NOLEN;
            commandArray[5] = KONST.UPLOAD_DATA_NOLEN;
            bool result = writeUSB(commandArray);
            if (result)
            {
                result = readUSB();
            }
            return result;
        }

        public static bool GetUpload()
        {
            return readUSB();
        }           

        public static bool UploadData()
        {
            byte[] commandArray = new byte[1];
            commandArray[0] = KONST.UPLOAD_DATA;
            bool result = writeUSB(commandArray);
            if (result)
            {
                result = readUSB();
            }
            return result;
        }

        public static bool UploadDataNoLen()
        {
            byte[] commandArray = new byte[1];
            commandArray[0] = KONST.UPLOAD_DATA_NOLEN;
            bool result = writeUSB(commandArray);
            if (result)
            {
                result = readUSB();
            }
            return result;
        }

        public static bool RunScript(int script, int repetitions)
        {
            // IMPORTANT NOTE: THIS ALWAYS CLEARS THE UPLOAD BUFFER FIRST!

            byte[] commandArray = new byte[4];
            commandArray[0] = KONST.CLR_UPLOAD_BUFFER;
            commandArray[1] = KONST.RUN_SCRIPT;
            commandArray[2] = (byte)script;
            commandArray[3] = (byte)repetitions;
            return writeUSB(commandArray);
        }  
        
        public static int DataClrAndDownload(byte[] dataArray, int startIndex)
        // returns index of next byte to be transmitted. 0 = failed
        {
            if (startIndex >= dataArray.Length)
            {
                return 0;
            }
            int length = dataArray.Length - startIndex;
            if (length > 61)
            {
                length = 61;
            }
            byte[] commandArray = new byte[3 + length];
            commandArray[0] = KONST.CLR_DOWNLOAD_BUFFER;
            commandArray[1] = KONST.DOWNLOAD_DATA;
            commandArray[2] = (byte)(length & 0xFF);
            for (int i = 0; i < length; i++)
            {
                commandArray[3 + i] = dataArray[startIndex + i];
            }
            if (writeUSB(commandArray))
            {
                return (startIndex + length);
            }
            else
            {
                return 0;
            }
        }

        public static int DataDownload(byte[] dataArray, int startIndex)
        // returns index of next byte to be transmitted. 0 = failed
        {
            if (startIndex >= dataArray.Length)
            {
                return 0;
            }
            int length = dataArray.Length - startIndex;
            if (length > 62)
            {
                length = 62;
            }
            byte[] commandArray = new byte[2 + length];
            commandArray[0] = KONST.DOWNLOAD_DATA;
            commandArray[1] = (byte)(length & 0xFF);
            for (int i = 0; i < length; i++)
            {
                commandArray[2 + i] = dataArray[startIndex + i];
            }
            if (writeUSB(commandArray))
            {
                return (startIndex + length);
            }
            else
            {
                return 0;
            }
        }      
        
        public static bool DownloadAddress3(int address)    
        {
            byte[] commandArray = new byte[6];
            commandArray[0] = KONST.CLR_DOWNLOAD_BUFFER;
            commandArray[1] = KONST.DOWNLOAD_DATA;
            commandArray[2] = 3;
            commandArray[3] = (byte)(address & 0xFF);
            commandArray[4] = (byte)(0xFF & (address >> 8));
            commandArray[5] = (byte)(0xFF & (address >> 16));
            return writeUSB(commandArray);
        }
        
        public static uint ComputeChecksum()
        {
            uint checksum = 0;
            
            if (DevFile.Families[GetActiveFamily()].BlankValue < 0xFFFF)
            { // 16F and baseline parts are caculated a word at a time.
                // prog mem first
                int progMemEnd = (int)DevFile.PartsList[ActivePart].ProgramMem;
                
                if (DevFile.PartsList[ActivePart].OSSCALSave)
                { // don't include last location for devices with OSSCAL 
                    progMemEnd--;
                }

                for (int idx = 0; idx < progMemEnd; idx++)
                {
                    checksum += DeviceBuffers.ProgramMemory[idx];
                }
                
                // config words
                for (int idx = 0; idx < DevFile.PartsList[ActivePart].ConfigWords; idx++)
                {
                    checksum += (DeviceBuffers.ConfigWords[idx] & DevFile.PartsList[ActivePart].ConfigMasks[idx]);
                }
                
                return (checksum & 0xFFFF);
            }
            else
            { //PIC18, PIC24 are computed a byte at a time.
                int progMemEnd = (int)DevFile.PartsList[ActivePart].ConfigAddr / DevFile.Families[GetActiveFamily()].ProgMemHexBytes;
                if (progMemEnd > DevFile.PartsList[ActivePart].ProgramMem)
                {
                    progMemEnd = (int)DevFile.PartsList[ActivePart].ProgramMem;
                }
                
                for (int idx = 0; idx < progMemEnd; idx++)
                {
                    uint memWord = DeviceBuffers.ProgramMemory[idx];
                    checksum += (memWord & 0x000000FF);
                    for (int bite = 1; bite < DevFile.Families[GetActiveFamily()].BytesPerLocation; bite++)
                    {
                        memWord >>= 8;
                        checksum += (memWord & 0x000000FF);
                    }    
                }

                // config words
                for (int idx = 0; idx < DevFile.PartsList[ActivePart].ConfigWords; idx++)
                {
                    uint memWord = (DeviceBuffers.ConfigWords[idx] & DevFile.PartsList[ActivePart].ConfigMasks[idx]);
                    checksum += (memWord & 0x000000FF);
                    checksum += ((memWord >> 8) & 0x000000FF);
                }
                return (checksum & 0xFFFF);
            }
            
        }

        public static void ResetBuffers()
        {
            DeviceBuffers = new DeviceData(DevFile.PartsList[ActivePart].ProgramMem,
                                            DevFile.PartsList[ActivePart].EEMem,
                                            DevFile.PartsList[ActivePart].ConfigWords,
                                            DevFile.PartsList[ActivePart].UserIDWords,
                                            DevFile.Families[GetActiveFamily()].BlankValue,
                                            DevFile.Families[GetActiveFamily()].EEMemAddressIncrement,
                                            DevFile.Families[GetActiveFamily()].UserIDBytes,
                                            DevFile.PartsList[ActivePart].ConfigBlank);
        }           
        
        public static void PrepNewPart()
        {
            ResetBuffers();
            setVppVoltage(DevFile.Families[GetActiveFamily()].Vpp);
            downloadPartScripts(GetActiveFamily());
        }   
        
        public static uint ReadDebugVector()
        {
            RunScript(KONST.PROG_ENTRY, 1);
            executeScript(DevFile.PartsList[ActivePart].DebugReadVectorScript);
            UploadData();
            RunScript(KONST.PROG_EXIT, 1);
            int configWords = 2;
            int bufferIndex = 2;                    // report starts on index 1, which is #bytes uploaded.
            uint returnWords = 0;
            for (int word = 0; word < configWords; word++)
            {
                uint config = (uint)Usb_read_array[bufferIndex++];
                config |= (uint)Usb_read_array[bufferIndex++] << 8;
                if (DevFile.Families[GetActiveFamily()].ProgMemShift > 0)
                {
                    config = (config >> 1) & DevFile.Families[GetActiveFamily()].BlankValue;
                }
                if (word == 0)
                    returnWords = config;
                else
                    returnWords += (config <<16);
            }     
            
            return returnWords;
        }

        public static void WriteDebugVector(uint debugWords)
        {
            int configWords = 2;
            byte[] configBuffer = new byte[4];

            RunScript(KONST.PROG_ENTRY, 1);

            for (int i = 0, j = 0; i < configWords; i++)
            {
                uint configWord = 0;
                if (i==0)
                    configWord = (debugWords & 0xFFFF);
                else
                    configWord = (debugWords >> 16);
                 if (DevFile.Families[GetActiveFamily()].ProgMemShift > 0)
                {
                    configWord = configWord << 1;
                }
                configBuffer[j++] = (byte)(configWord & 0xFF);
                configBuffer[j++] = (byte)((configWord >> 8) & 0xFF);
            }
            DataClrAndDownload(configBuffer, 0);
            executeScript(DevFile.PartsList[ActivePart].DebugWriteVectorScript);
            RunScript(KONST.PROG_EXIT, 1);

        }        
                
        
        // ================================== PRIVATE METHODS ========================================

        private static bool setVppVoltage(float voltage)
        {
            byte ccpValue = 0x40;
            byte vppADC = (byte)(voltage * 18.61F);
            byte vFault = (byte)(0.7F * voltage * 18.61F);

            byte[] commandArray = new byte[4];
            commandArray[0] = KONST.SETVPP;
            commandArray[1] = ccpValue;
            commandArray[2] = vppADC;
            commandArray[3] = vFault;
            return writeUSB(commandArray);
        }  

        private static bool readPICkitVoltages(ref float vdd, ref float vpp)
        {
            byte[] commandArray = new byte[1];
            commandArray[0] = KONST.READ_VOLTAGES;
            if (writeUSB(commandArray))
            {
                if (readUSB())
                {
                    float valueADC = (float)((Usb_read_array[2] * 256) + Usb_read_array[1]);
                    vdd = (valueADC / 65536) * 5.0F;
                    valueADC = (float)((Usb_read_array[4] * 256) + Usb_read_array[3]);
                    vpp = (valueADC / 65536) * 13.7F;
                    return true;
                }
            }
            return false;
        }

        private static ushort readPkStatus()
        {
            byte[] commandArray = new byte[1];
            commandArray[0] = KONST.READ_STATUS;
            if (writeUSB(commandArray))
            {
                if (readUSB())
                {
                    return (ushort)(Usb_read_array[2] * 256 + Usb_read_array[1]);
                }
                return 0xFFFF;
            }

            return 0xFFFF;
        }        
        
        private static bool writeUSB(byte[] commandList)
        {
            int bytesWritten = 0;

            Usb_write_array[0] = 0;                         // first byte must always be zero.        
            for (int index = 1; index < Usb_write_array.Length; index++)
            {
                Usb_write_array[index] = KONST.END_OF_BUFFER;              // init array to all END_OF_BUFFER cmds.
            }
            Array.Copy(commandList, 0, Usb_write_array, 1, commandList.Length);
            bool writeResult = USB.WriteFile(usbWriteHandle, Usb_write_array, Usb_write_array.Length, ref bytesWritten, 0);
            if (bytesWritten != Usb_write_array.Length)
            {
                return false;
            }
            return writeResult;
        
        }
        
        private static bool readUSB()
        {
            int bytesRead = 0;
        
            bool readResult = USB.ReadFile(usbReadHandle, Usb_read_array, Usb_read_array.Length, ref bytesRead, 0);
            if (bytesRead != Usb_read_array.Length)
            {
                return false;
            }
            return readResult;            
            
        }


        private static bool searchDevice(int familyIndex, bool resetOnNoDevice, bool keepVddOn)
        {
            int lastPart = ActivePart;  // remember the current part
            if (ActivePart != 0)
            {
                lastFoundPart = ActivePart;
            }
            
            // NOTE: the interface portion should ensure that self-powered targets
            // are detected before calling this function.

            // Set VPP voltage by family
            float vpp = DevFile.Families[familyIndex].Vpp;
            if (vpp < 1)  
            { // When nominally zero, use VDD voltage
                setVppVoltage(vddLastSet);       
            }
            else
            {
                setVppVoltage(vpp);
            }
            
            // Turn on Vdd (if self-powered, just turns off ground resistor)
            VddOn();

            // use direct execute scripts when checking for a part
            executeScript(DevFile.Families[familyIndex].ProgEntryScript);
            executeScript(DevFile.Families[familyIndex].ReadDevIDScript);
            UploadData();
            executeScript(DevFile.Families[familyIndex].ProgExitScript);
            
            // Turn off Vdd (if PICkit-powered, turns on ground resistor)
            if (!keepVddOn)
            { // don't want it off when user wants PICkit 2 VDD "ON"
                VddOff();
            }

            uint deviceID = (uint)(Usb_read_array[3] * 256 + Usb_read_array[2]);
            for (int shift = 0; shift < DevFile.Families[familyIndex].ProgMemShift; shift++)
            {
                deviceID >>= 1;         // midrange/baseline part results must be shifted by 1
            }
            int rev = (int)(deviceID & ~DevFile.Families[familyIndex].DeviceIDMask);
            deviceID &= DevFile.Families[familyIndex].DeviceIDMask; // mask off version bits.

            LastDeviceID = deviceID;

            // Search through the device file to see if we find the part
            ActivePart = 0; // no device is default.
            for (int partEntry = 0; partEntry < DevFile.PartsList.Length; partEntry++)
            {
                if (DevFile.PartsList[partEntry].Family == familyIndex)
                { // don't check device ID if in a different family
                    if (DevFile.PartsList[partEntry].DeviceID == deviceID)
                    {
                        ActivePart = partEntry;
                        ActiveRev = rev;
                        break;  // found a match - get out of the loop.
                    }
                }
            }

            if (ActivePart == 0) // not a known part
            {   // still need a buffer object in existance.
                if (lastPart != 0)
                {
                    DevFile.PartsList[ActivePart] = DevFile.PartsList[lastPart];
                    DevFile.PartsList[ActivePart].DeviceID = 0;
                    DevFile.PartsList[ActivePart].PartName = "Unsupported Part";
                }
                if (resetOnNoDevice)
                {
                    ResetBuffers();
                }
                return false;   // we're done
            }

            //if ((ActivePart == lastPart) && (scriptBufferChecksum == getScriptBufferChecksum())) 
            //{// same part we have been using (ensure scipt buffer hasn't been corrupted.)
            //    return true;    // don't need to download scripts as they should already be there.
            //}
            
            if ((ActivePart == lastFoundPart) && (scriptBufferChecksum == getScriptBufferChecksum()))
            {// same as the last part we were connected to.
                return true;    // don't need to download scripts as they should already be there.
            }

            // Getting here means we've found a part, but it's a new one so we need to download scripts
            downloadPartScripts(familyIndex);

            // create a new set of device buffers
            ResetBuffers();

            // Get OSCCAL if exists
            if (DevFile.PartsList[ActivePart].OSSCALSave)
            {
                VddOn();
                ReadOSSCAL();
            }
            if (DevFile.PartsList[ActivePart].BandGapMask > 0)
            {
                VddOn();
                ReadBandGap();
            }
            if (!keepVddOn)
            {
                VddOff();
            }

            return true;

        }
  
        
        private static void downloadPartScripts(int familyIndex)
        {
            byte[] commandArray = new byte[1];
            commandArray[0] = KONST.CLR_SCRIPT_BUFFER;      // clear script buffer- we're loading new scripts
            bool result = writeUSB(commandArray);

            // program entry
            if (DevFile.Families[familyIndex].ProgEntryScript != 0) // don't download non-existant scripts
            {
                downloadScript(KONST.PROG_ENTRY, DevFile.Families[familyIndex].ProgEntryScript);
            }
            // program exit
            if (DevFile.Families[familyIndex].ProgExitScript != 0) // don't download non-existant scripts
            {
                downloadScript(KONST.PROG_EXIT, DevFile.Families[familyIndex].ProgExitScript);
            }
            // read device id
            if (DevFile.Families[familyIndex].ReadDevIDScript != 0) // don't download non-existant scripts
            {
                downloadScript(KONST.RD_DEVID, DevFile.Families[familyIndex].ReadDevIDScript);
            }
            // read program memory
            if (DevFile.PartsList[ActivePart].ProgMemRdScript != 0) // don't download non-existant scripts
            {
                downloadScript(KONST.PROGMEM_RD, DevFile.PartsList[ActivePart].ProgMemRdScript);
            }
            // set program memory address
            if (DevFile.PartsList[ActivePart].ProgMemAddrSetScript != 0) // don't download non-existant scripts
            {
                downloadScript(KONST.PROGMEM_ADDRSET, DevFile.PartsList[ActivePart].ProgMemAddrSetScript);
            }
            // prepare for program memory write
            if (DevFile.PartsList[ActivePart].ProgMemWrPrepScript != 0) // don't download non-existant scripts
            {
                downloadScript(KONST.PROGMEM_WR_PREP, DevFile.PartsList[ActivePart].ProgMemWrPrepScript);
            }
            // program memory write                 
            if (DevFile.PartsList[ActivePart].ProgMemWrScript != 0) // don't download non-existant scripts
            {
                downloadScript(KONST.PROGMEM_WR, DevFile.PartsList[ActivePart].ProgMemWrScript);
            }
            // prep for ee read               
            if (DevFile.PartsList[ActivePart].EERdPrepScript != 0) // don't download non-existant scripts
            {
                downloadScript(KONST.EE_RD_PREP, DevFile.PartsList[ActivePart].EERdPrepScript);
            }
            // ee read               
            if (DevFile.PartsList[ActivePart].EERdScript != 0) // don't download non-existant scripts
            {
                downloadScript(KONST.EE_RD, DevFile.PartsList[ActivePart].EERdScript);
            }
            // prep for ee write               
            if (DevFile.PartsList[ActivePart].EEWrPrepScript != 0) // don't download non-existant scripts
            {
                downloadScript(KONST.EE_WR_PREP, DevFile.PartsList[ActivePart].EEWrPrepScript);
            }
            // ee write               
            if (DevFile.PartsList[ActivePart].EEWrScript != 0) // don't download non-existant scripts
            {
                downloadScript(KONST.EE_WR, DevFile.PartsList[ActivePart].EEWrScript);
            }
            // prep for config read       
            if (DevFile.PartsList[ActivePart].ConfigRdPrepScript != 0) // don't download non-existant scripts
            {
                downloadScript(KONST.CONFIG_RD_PREP, DevFile.PartsList[ActivePart].ConfigRdPrepScript);
            }
            // config read       
            if (DevFile.PartsList[ActivePart].ConfigRdScript != 0) // don't download non-existant scripts
            {
                downloadScript(KONST.CONFIG_RD, DevFile.PartsList[ActivePart].ConfigRdScript);
            }
            // prep for config write       
            if (DevFile.PartsList[ActivePart].ConfigWrPrepScript != 0) // don't download non-existant scripts
            {
                downloadScript(KONST.CONFIG_WR_PREP, DevFile.PartsList[ActivePart].ConfigWrPrepScript);
            }
            // config write       
            if (DevFile.PartsList[ActivePart].ConfigWrScript != 0) // don't download non-existant scripts
            {
                downloadScript(KONST.CONFIG_WR, DevFile.PartsList[ActivePart].ConfigWrScript);
            }
            // prep for user id read      
            if (DevFile.PartsList[ActivePart].UserIDRdPrepScript != 0) // don't download non-existant scripts
            {
                downloadScript(KONST.USERID_RD_PREP, DevFile.PartsList[ActivePart].UserIDRdPrepScript);
            }
            // user id read      
            if (DevFile.PartsList[ActivePart].UserIDRdScript != 0) // don't download non-existant scripts
            {
                downloadScript(KONST.USERID_RD, DevFile.PartsList[ActivePart].UserIDRdScript);
            }
            // prep for user id write      
            if (DevFile.PartsList[ActivePart].UserIDWrPrepScript != 0) // don't download non-existant scripts
            {
                downloadScript(KONST.USERID_WR_PREP, DevFile.PartsList[ActivePart].UserIDWrPrepScript);
            }
            // user id write      
            if (DevFile.PartsList[ActivePart].UserIDWrScript != 0) // don't download non-existant scripts
            {
                downloadScript(KONST.USERID_WR, DevFile.PartsList[ActivePart].UserIDWrScript);
            }
            // read osscal      
            if (DevFile.PartsList[ActivePart].OSCCALRdScript != 0) // don't download non-existant scripts
            {
                downloadScript(KONST.OSSCAL_RD, DevFile.PartsList[ActivePart].OSCCALRdScript);
            }
            // write osscal      
            if (DevFile.PartsList[ActivePart].OSCCALWrScript != 0) // don't download non-existant scripts
            {
                downloadScript(KONST.OSSCAL_WR, DevFile.PartsList[ActivePart].OSCCALWrScript);
            }
            // chip erase      
            if (DevFile.PartsList[ActivePart].ChipEraseScript != 0) // don't download non-existant scripts
            {
                downloadScript(KONST.ERASE_CHIP, DevFile.PartsList[ActivePart].ChipEraseScript);
            }
            // program memory erase 
            if (DevFile.PartsList[ActivePart].ProgMemEraseScript != 0) // don't download non-existant scripts
            {
                downloadScript(KONST.ERASE_PROGMEM, DevFile.PartsList[ActivePart].ProgMemEraseScript);
            }
            // ee erase 
            if (DevFile.PartsList[ActivePart].EEMemEraseScript != 0) // don't download non-existant scripts
            {
                downloadScript(KONST.ERASE_EE, DevFile.PartsList[ActivePart].EEMemEraseScript);
            }
            // config memory erase
            if (DevFile.PartsList[ActivePart].ConfigMemEraseScript != 0) // don't download non-existant scripts
            {
                downloadScript(KONST.ERASE_CONFIG, DevFile.PartsList[ActivePart].ConfigMemEraseScript);
            }
            
            // get script buffer checksum
            scriptBufferChecksum = getScriptBufferChecksum();
        }
        
        private static uint getScriptBufferChecksum()
        {
            byte[] commandArray = new byte[1];
            commandArray[0] = KONST.SCRIPT_BUFFER_CHKSUM;
            if (writeUSB(commandArray))
            {
                if (readUSB())
                {
                    uint checksum = (uint)Usb_read_array[4];
                    checksum += (uint)(Usb_read_array[3] << 8);
                    checksum += (uint)(Usb_read_array[2] << 16);
                    checksum += (uint)(Usb_read_array[1] << 24);
                    
                    return checksum;
                }
                return 0;
            }
            return 0;        
        }

        // Louis - upload an arbitrary script (for downloading)
        // It seems that we must expose the inner behavior of the PICkit 2.
        public static bool ExecuteScript(byte[] script)
        {
            byte[] commandArray = new byte[3 + script.Length];
            commandArray[0] = KONST.CLR_UPLOAD_BUFFER;
            commandArray[1] = KONST.EXECUTE_SCRIPT;
            commandArray[2] = (byte)script.Length;
            for (int n = 0; n < script.Length; n++)
            {
                commandArray[3 + n] = script[n];
            }
            return writeUSB(commandArray);
        }

        private static bool executeScript(int scriptArrayIndex)
        {
            // IMPORTANT NOTE: THIS ALWAYS CLEARS THE UPLOAD BUFFER FIRST!
            int scriptLength = DevFile.Scripts[--scriptArrayIndex].ScriptLength;

            byte[] commandArray = new byte[3 + scriptLength];
            commandArray[0] = KONST.CLR_UPLOAD_BUFFER;
            commandArray[1] = KONST.EXECUTE_SCRIPT;
            commandArray[2] = (byte)scriptLength;
            for (int n = 0; n < scriptLength; n++)
            {
                commandArray[3 + n] = (byte)DevFile.Scripts[scriptArrayIndex].Script[n];
            }
            return writeUSB(commandArray);
        }      

        private static bool downloadScript(byte scriptBufferLocation, int scriptArrayIndex)
        {
            int scriptLength = DevFile.Scripts[--scriptArrayIndex].ScriptLength;

            byte[] commandArray = new byte[3 + scriptLength];
            commandArray[0] = KONST.DOWNLOAD_SCRIPT;
            commandArray[1] = scriptBufferLocation;
            commandArray[2] = (byte)scriptLength;
            for (int n = 0; n < scriptLength; n++)
            {
                commandArray[3 + n] = (byte)DevFile.Scripts[scriptArrayIndex].Script[n];
            }
            return writeUSB(commandArray);
        }

      
                   
    
    }
}
